clear; close all; clc
set(0,'DefaultFigureWindowStyle','docked')
%% Localization with particle filter
vid = 0;

%% Generate map
%GenMap();
load('map.mat')
fig_c = figure('visible','off');
figure(2)
subplot(2,1,1)
plot(map)

% An anonimous function to get the altitude data
% This is the heasurement function
h = @(x) map(round(min(1000,max(1,x))));

%% Model of tha airplane
% Maintains the altitude and flies with CV
% We do not know the position and the velocity

Ts = 0.5; % [sec]
v=50; % [m/s] a vals sebessg

% Starting position
x(1) = 10;
z(1) = h(x(1));

%% Find the airplane! What se should do:
% 1. Wait for the first measurement: z(1)
% 2. Invert the measurement equation: x=h^-1(z)
% 3. No we have an initial position: x(1), start the filtering process
% 4. Make a prediction based on the motion model
% 5. Do the update when measurement arrives
% 6. goto 4.

% z_k = h(X_k) is the measurement model. We cannot invert this! It is not
% even in analytic form, it is a dataset. => Use particle filter

% The motion model is corrupted by AWGN: X_k+1 = F * X_k + w
% w is zero mean and the covariance is
Qm=[4, 0; 0, 4]; % []

% The state vector contains the position and speed values: X = [x; v]
% The system matrix describing a CV model in discrete time is:
F=[1 Ts; 0 1];

%% Particle filter
% No need to invert the measurement equation

% The particles are realisations of the state vector: P(i).x, P(i).v 
Np=1000; % Number of particles used

% Initially the airplane can be anywhere: we throw particles all over the
% map. The x coordinate can have values between 1...1000 m
% The speed is in -100...100 m/s
% all particles have the same initial weight

% Create initial particles
for i=1:Np
    P(1).x(i) = rand * 999 + 1;
    P(1).v(i) = 200 * rand - 100;
    P(1).w(i) = 1 / Np;
end
%subplot(2,1,2)

% The first measurement arrives: z(1)
% We evaluate the measurement model on all particles and compare it with
% z(1)
% The particles will get weights based on how close they are to z(1)

% The uncertainty on the measurement model is modeld by AWGN: z = h(x) + v
% with zero mean and std=2 m
Qz=4; % [m^2]

% Bootstrap particle filter: a proposal distribution = state transition
% density, so the particle weights are the likelihoods
% the likelihood is a Gaussian (beacure of the measurement model)
% g(z|x) = N(z; h(x),Q)
% evaluate g(z|x) for every particle to get the likelihoods (weights)
% the weights should be normalized

% Calculate weights
P(1).w = normpdf(h(P(1).x),z(1),Qz);
% Normalize
P(1).w = P(1).w / sum(P(1).w);

% resample: from the set P(1).x , Np instances, with replacement, with
% P(1).w weights
P(1).x = randsample(P(1).x, Np, true, P(1).w);

% Plot the airplane and the measurement
hold on
plot(x(1),200,'rs')
plot([1,1000],[h(x(1)) h(x(1))],'--r')
% plot([1,1000],[h(x(1)) h(x(1))]-Qz^0.5,'--r')
% plot([1,1000],[h(x(1)) h(x(1))]+Qz^0.5,'--r')
xlim([0,1000])
ylim([0,300])

subplot(2,1,2)
scatter(P(1).x,200*ones(1,Np))
xlim([0,1000])
ylim([0,300])

%% Next step
% Move the airplane and generate new measurement
x(2) = x(1) + Ts * v;
z(2) = h(x(2));

% Prediction for the state
for i=1:Np
    X = F * [P(1).x(i);P(1).v(i)] + (mvnrnd([0,0],Qm))';
    P(2).x(i) = X(1);
    P(2).v(i) = X(2);
end

subplot(2,2,3)
scatter(P(2).x,200*ones(1,Np))
xlim([0,1000])
ylim([0,300])

% likelihoods
P(2).w = normpdf(h(P(2).x),z(2),Qz);
P(2).w = P(2).w / sum(P(2).w);
P(2).x = randsample(P(2).x,Np,true,P(2).w);

subplot(2,2,1)
plot(x(2),200,'rs')
plot([1,1000],[h(x(2)) h(x(2))],'--r')

subplot(2,2,3)
scatter(P(2).x,200*ones(1,Np))
xlim([0,1000])
ylim([0,300])
drawnow

% Create video file
if vid==1
    vidobj = VideoWriter('2.mp4','MPEG-4')
    open(vidobj);
end

%% Do in cycle
Nk = round(1000/v/Ts)
err = zeros(1,Nk);

% First 2 steps already done
for k=3:Nk-1
    
    % Simulation
    x(k) = x(k-1) + Ts*v;
    z(k) = h(x(k));
    
    % Prediction
    for i=1:Np
        X = F * [P(k-1).x(i);P(k-1).v(i)] + (mvnrnd([0,0],Qm))';
        P(k).x(i)=X(1);
        P(k).v(i)=X(2);
    end
    
    figure(2)
    subplot(2,2,3)
    scatter(P(2).x,200*ones(1,Np))
    xlim([0,1000])
    ylim([0,300])
    
    % Update
    P(k).w=normpdf(h(P(k).x),z(k),Qz);
    P(k).w = P(k).w / sum(P(k).w);
    
    % Point estimate: MAP - maximum a posterior
    % In multimodal distribution the expected value does not make sense
    % MAP: the particle with the greatest probability will be the estimate
    % that is the particle with the greatest weight
    [~, idx] = max(P(k).w);
    P(k).estx = P(k).x(idx);
    P(k).estv = P(k).v(idx);
    
    % Resample
    P(k).x = randsample(P(k).x,Np,true,P(k).w);
    err(k) = x(k) - P(k).estx;
    
    % Regularize
%     randmat = randn(2,Np);
%     C=cov([P(k).x;P(k).v]');
%     C=sqrtm(C);
%     A=(4/(2+2))^(1/(2+4));
%     hh=A*Np^(-1/(2+4));
%     XX=[P(k).x;P(k).v] + 0.5*hh*C*randmat;
%     P(k).x = XX(1,:);
%     P(k).v = XX(2,:);
    
    subplot(2,2,1)
    hold off
    plot(map)
    hold on
    plot(x(k),200,'rs')
    plot([1,1000],[h(x(k)) h(x(k))],'--r')
    xlim([0,1000])
    ylim([0,300])
    
    subplot(2,2,3)
    hold off
    plot(map)
    hold on
    scatter(P(k).x,200*ones(1,Np),[],'b')
    scatter(P(k).estx,200,[],'r','filled')
    xlim([0,1000])
    ylim([0,300])
    %pause(0.2)
    
    % Contourplot in the state space
    [xel xc] = hist(P(k).x,1:1000);
    [vel vc] = hist(P(k).v,1:100);

    Xel = repmat(xel,100,1);
    Vel = repmat(vel',1,1000);

    
    M = Xel + Vel;
    subplot(2,2,2)
    %set(gcf,'visible','off')
    %mov(k)=getframe(gca);
    
    if vid==1
        contourf(M,25)
        colormap(summer)
    
        f=getframe(gca);
        writeVideo(vidobj,f);
    end
end

if vid==1
    close(vidobj);

    vidobj = VideoReader('2.mp4');
    vidWidth = vidobj.Width;
    vidHeight = vidobj.Height;
    mov = struct('cdata',zeros(vidHeight,vidWidth,3,'uint8'),'colormap',[]);
    
    k=1;
    while hasFrame(vidobj)
        mov(k).cdata=readFrame(vidobj);
        k = k+1;
    end

    subplot(2,2,2)
    movie(gca,mov,1,vidobj.FrameRate);
end

%% Evaluate estimation
for i=1:Nk-1
    V(i)=mean(P(i).x);
end
mean(V)

subplot(2,2,4)
plot(err)
title('Estimation error')
xlabel('Time')
ylabel('Position')

license('inuse')